@@ -2,14 +2,15 @@ from __future__ import print_function
 """degrotesque.py
 
 A tiny web type setter.
+Version 1.6
 
-(c) Daniel Krajzewicz 2020-2021
+(c) Daniel Krajzewicz 2020-2022
 daniel@krajzewicz.de
 http://www.krajzewicz.de
 https://github.com/dkrajzew/degrotesque
 http://www.krajzewicz.de/blog/degrotesque.php
 
-Available under LGPL 3.0 or later, all rights reserved
+Available under LGPL 3 or later, all rights reserved
 """
 
 
@@ -18,94 +19,98 @@ import sys, glob, os, io, shutil, re
 from optparse import OptionParser
 from html.parser import HTMLParser
 
+
+
 # --- variables and constants ---------------------------------------
 """A database of actions"""
 actionsDB = {
  # english quotes
  "quotes.english": [
-   [[u"(\\s+)'", u"'"], [u"\\1&lsquo;", u"&rsquo;"]],
-   [[u"\"", u"\""], [u"&ldquo;", u"&rdquo;"]]
+   [[u"(\\s+)'", u"'"],             [u"\\1&lsquo;", u"&rsquo;"],    [u"\\1&#8216;", u"&#8217;"]],
+   [[u"\"", u"\""],                 [u"&ldquo;", u"&rdquo;"],       [u"&#8220;", u"&#8221;"]]
   ],
 
  # french quotes
  "quotes.french": [
-   [[u"&lt;&lt;", u"&gt;&gt;"], [u"&laquo;", u"&raquo;"]],
-   [[u"&lt;", u"&gt;"], [u"&lsaquo;", u"&rsaquo;"]]
+   [[u"&lt;&lt;", u"&gt;&gt;"],     [u"&laquo;", u"&raquo;"],       [u"&#x00AB;", u"&#x00BB;"]],
+   [[u"&lt;", u"&gt;"],             [u"&lsaquo;", u"&rsaquo;"],     [u"&#x2039;", u"&#x203A;"]]
   ],
   
  # german quotes
  "quotes.german": [
-   [[u"(\\s+)'", u"'"], [u"\\1&sbquo;", u"&rsquo;"]],
-   [[u"\"", u"\""], [u"&bdquo;", u"&rdquo;"]]
+   [[u"(\\s+)'", u"'"],             [u"\\1&sbquo;", u"&rsquo;"],    [u"\\1&#x201A;", u"&#x2019;"]],
+   [[u"\"", u"\""],                 [u"&bdquo;", u"&rdquo;"],       [u"&#x201E;", u"&#x201D;"]]
   ],
   
  # conversion to HTML quotes (<q>)
  "to_quotes": [
-   [[u"(\\s+)'", u"'"], [u"\\1<q>", u"</q>"]],
-   [[u"\"", u"\""], [u"<q>", u"</q>"]],
-   [[u"&lt;&lt;", u"&gt;&gt;"], [u"<q>", u"</q>"]],
-   [[u"&lt;", u"&gt;"], [u"<q>", u"</q>"]]
+   [[u"(\\s+)'", u"'"],             [u"\\1<q>", u"</q>"],           [u"\\1<q>", u"</q>"]],
+   [[u"\"", u"\""],                 [u"<q>", u"</q>"],              [u"<q>", u"</q>"]],
+   [[u"&lt;&lt;", u"&gt;&gt;"],     [u"<q>", u"</q>"],              [u"<q>", u"</q>"]],
+   [[u"&lt;", u"&gt;"],             [u"<q>", u"</q>"],              [u"<q>", u"</q>"]]
   ],
   
  # commercial signs
  "commercial": [
-   [[u"\\([c|C]\\)", None], [u"&copy;", None]],
-   [[u"\\([r|R]\\)", None], [u"&reg;", None]],
-   [[u"\\([t|T][m|M]\\)", None], [u"&trade;", None]]
+   [[u"\\([c|C]\\)", None],         [u"&copy;", None],              [u"&#169;", None]],
+   [[u"\\([r|R]\\)", None],         [u"&reg;", None],               [u"&#174;", None]],
+   [[u"\\([t|T][m|M]\\)", None],    [u"&trade;", None],             [u"&#8482;", None]]
   ],
   
  # dashes
  "dashes": [
    # missing: ndash for number ranges 
-   [[u"(\\s+)-(\\s+)", None], [u"\\1&mdash;\\2", None]],
-   [[u"([\\d]+)-([\\d]+)", None], [u"\\1&ndash;\\2", None]]
-   #[[u"\\W-([\\d]+)", None], [u"&ndash;\\1", None]],
-   #[[u"([\\d]+)-\\W", None], [u"\\1&ndash;", None]],
+   [[u"(\\s+)-(\\s+)", None],       [u"\\1&mdash;\\2", None],       [u"\\1&#8212;\\2", None]],
+   [[u"([\\d]+)-([\\d]+)", None],   [u"\\1&ndash;\\2", None],       [u"\\1&#8211;\\2", None]]
   ],
 
  # bullets
  "bullets": [
-   [[u"\\*", None], [u"&bull;", None]]
+   [[u"\\*", None],                 [u"&bull;", None],              [u"&#8226;", None]]
   ],
     
  # ellipsis
  "ellipsis": [
-   [[u"\\.\\.\\.", None], [u"&hellip;", None]]
+   [[u"\\.\\.\\.", None],           [u"&hellip;", None],            [u"&#8230;", None]]
   ],
     
  # apostrophe
  "apostrophe": [
-   [[u"'", None], [u"&apos;", None]]
-  ],
-    
- # math signs
- "math": [
-   # [[""], ["", "&deg;", ""]],
-   [[u"\\+/-", None], [u"&plusmn;", None]],
-   [[u"1/2", None], [u"&frac12;", None]],
-   [[u"1/4", None], [u"&frac14;", None]],
-   [[u"3/4", None], [u"&frac34;", None]],
-   [[u"\\~", None], [u"&asymp;", None]],
-   [[u"\\!=", None], [u"&ne;", None]],
-   [[u"&lt;=", None], [u"&le;", None]],
-   [[u"&gt;=", None], [u"&ge;", None]],
-   [[u"([\\d]+)(\\s*)\*(\\s*)([\\d]+)", None], [u"\\1\\2&times;\\3\\4", None]],
-   [[u"([\\d]+)(\\s*)x(\\s*)([\\d]+)", None], [u"\\1\\2&times;\\3\\4", None]],
-   [[u"([\\d]+)(\\s*)/(\\s*)([\\d]+)", None], [u"\\1\\2&divide;\\3\\4", None]]
+   [[u"'", None],                   [u"&apos;", None],              [u"&#39;", None]]
   ],
     
  # dagger
  "dagger": [
-   [[u"\\*\\*", None], [u"&Dagger;", None]],
-   [[u"\\*", None], [u"&dagger;", None]]
+   [[u"\\*\\*", None],              [u"&Dagger;", None],            [u"&#8225;", None]],
+   [[u"\\*", None],                 [u"&dagger;", None],            [u"&#8224;", None]]
   ],    
 
+ # math signs
+ "math": [
+   [[u"\\+/-", None],               [u"&plusmn;", None],            [u"&#177;", None]],
+   [[u"1/2", None],                 [u"&frac12;", None],            [u"&#189;", None]],
+   [[u"1/4", None],                 [u"&frac14;", None],            [u"&#188;", None]],
+   [[u"3/4", None],                 [u"&frac34;", None],            [u"&#190;", None]],
+   [[u"\\~", None],                 [u"&asymp;", None],             [u"&#8776;", None]],
+   [[u"\\!=", None],                [u"&ne;", None],                [u"&#8800;", None]],
+   [[u"&lt;=", None],               [u"&le;", None],                [u"&#8804;", None]],
+   [[u"&gt;=", None],               [u"&ge;", None],                [u"&#8805;", None]],
+   [[u"([\\d]+)(\\s*)\*(\\s*)([\\d]+)", None],  [u"\\1\\2&times;\\3\\4", None],     [u"\\1\\2&#215;\\3\\4", None]],
+   [[u"([\\d]+)(\\s*)x(\\s*)([\\d]+)", None],   [u"\\1\\2&times;\\3\\4", None],     [u"\\1\\2&#215;\\3\\4", None]],
+   [[u"([\\d]+)(\\s*)/(\\s*)([\\d]+)", None],   [u"\\1\\2&divide;\\3\\4", None],    [u"\\1\\2&#247;\\3\\4", None]]
+  ],
+    
+ # chem
+ "chem": [
+   [[u"([a-zA-Z]+)([\\d]+)", None],        [u"\\1<sub>\\2</sub>", None],   [u"\\1<sub>\\2</sub>", None]]
+  ],
+    
  # masks
  "masks": [
-   [[u"978-([\\d]+)-([\\d]+)-([\\d]+)-([\\d])(\\D)", None], [u"978-\\1-\\2-\\3-\\4\\5", None]],
-   [[u"979-([\\d]+)-([\\d]+)-([\\d]+)-([\\d])(\\D)", None], [u"979-\\1-\\2-\\3-\\4\\5", None]],
-   [[u"([\\d]+)-([\\d]+)-([\\d]+)-([\\d])(\\D)", None], [u"\\1-\\2-\\3-\\4\\5", None]],
-   [[u"ISSN (\\d{4})-(\\d{4})(\\D)", None], [u"ISSN \\1-\\2\\3", None]]
+   [[u"978-([\\d]+)-([\\d]+)-([\\d]+)-([\\d])(\\D)", None], [u"978-\\1-\\2-\\3-\\4\\5", None],  [u"978-\\1-\\2-\\3-\\4\\5", None]],
+   [[u"979-([\\d]+)-([\\d]+)-([\\d]+)-([\\d])(\\D)", None], [u"979-\\1-\\2-\\3-\\4\\5", None],  [u"979-\\1-\\2-\\3-\\4\\5", None]],
+   [[u"([\\d]+)-([\\d]+)-([\\d]+)-([\\d])(\\D)", None],     [u"\\1-\\2-\\3-\\4\\5", None],      [u"\\1-\\2-\\3-\\4\\5", None]],
+   [[u"ISSN (\\d{4})-(\\d{4})(\\D)", None],                 [u"ISSN \\1-\\2\\3", None],         [u"ISSN \\1-\\2\\3", None]]
   ]    
 }
 
@@ -132,30 +137,27 @@ class Degrotesque():
      and elements to skip.
      Some internal methods exist for determining which parts of the document 
      shall processed and which ones are to skip."""
+     
   # --- init
   def __init__(self):
-     """Sets defaults for the elements which contents shall not be processed.
-        Sets defaults for actions to perform."""
-     # list of elements which contents shall not be processed
-     self._elementsToSkip = [
-       u"script", u"code", u"style", u"pre", u"?", u"?php", 
-       u"%", u"%=", u"%@", u"%--", u"%!",
-       u"!--"
-     ]
-     # the actions to apply
-     self.restoreDefaultActions()
+    """Sets defaults for the elements which contents shall not be processed.
+       Sets defaults for actions to perform."""
+    # the elements to skip
+    self._restoreDefaultElementsToSkip()
+    # the actions to apply
+    self._restoreDefaultActions()
      
      
   # --- restoreDefaultActions
-  def restoreDefaultActions(self):
-     """Instantiates default actions"""     
-     self._actions = []
-     self._actions.extend(actionsDB["masks"])
-     self._actions.extend(actionsDB["quotes.english"])
-     self._actions.extend(actionsDB["dashes"])
-     self._actions.extend(actionsDB["ellipsis"])
-     self._actions.extend(actionsDB["math"])
-     self._actions.extend(actionsDB["apostrophe"])
+  def _restoreDefaultActions(self):
+    """Instantiates default actions"""     
+    self._actions = []
+    self._actions.extend(actionsDB["masks"])
+    self._actions.extend(actionsDB["quotes.english"])
+    self._actions.extend(actionsDB["dashes"])
+    self._actions.extend(actionsDB["ellipsis"])
+    self._actions.extend(actionsDB["math"])
+    self._actions.extend(actionsDB["apostrophe"])
   
     
   # --- setActions
@@ -178,13 +180,26 @@ class Degrotesque():
         raise ValueError("Action '%s' is not known." % (an))
 
 
+  # --- restoreDefaultActions
+  def _restoreDefaultElementsToSkip(self):
+    """Instantiates default elements to skip"""
+    # list of elements which contents shall not be processed
+    self._elementsToSkip = [
+      u"script", u"code", u"style", u"pre", u"?", u"?php", 
+      u"%", u"%=", u"%@", u"%--", u"%!",
+      u"!--", "!DOCTYPE"
+    ]
+
+
   # --- setToSkip
   def setToSkip(self, toSkipNames):
     """Returns the elements which contents shall not be changed.
        If the given names of elements are None or empty, the default elements 
        to skip are used.
        Otherwise, a list with the elements to skip is built.
-       :param toSkipNames The names of elements which shall not be changed"""
+       :param toSkipNames The names of elements which shall not be changed
+       :todo Warn user if a non-XML-character occurs?
+       """
     if toSkipNames==None or len(toSkipNames)==0:
       return 
     self._elementsToSkip = [x.strip() for x in toSkipNames.split(',')] 
@@ -237,13 +252,19 @@ class Degrotesque():
       ib = i
       if tb=="?" or tb=="?php":
         # assumption: php stuff is always closed by ?>
-        ie = html.index("?>", ib)
+        ie = html.find("?>", ib)
+        if ie<0: raise ValueError("Unclosed '<%s' element at position %s." % (tb, i))
+        ie += 1
       elif tb=="%" or tb=="%=" or tb=="%@" or tb=="%--" or tb=="%!":
-        # assumption: jsp/asp stuff is always closed by ?>
-        ie = html.index("%>", ib)
+        # assumption: jsp/asp stuff is always closed by %>
+        ie = html.find("%>", ib)
+        if ie<0: raise ValueError("Unclosed '<%s' element at position %s." % (tb, i))
+        ie += 1
       elif tb=="!--":
-        # comments are always closed by a -->
-        ie = html.index("-->", ib)
+        # comments are always closed by -->
+        ie = html.find("-->", ib)
+        if ie<0: raise ValueError("Unclosed '<%s' element at position %s." % (tb, i))
+        ie += 2
       elif tb=="!DOCTYPE":
         # DOCTYPE: find matching >
         ie = ib+1
@@ -253,27 +274,24 @@ class Degrotesque():
           elif html[ie]==">": num = num + 1
           if num==0: break
           ie = ie + 1
+        ie -= 1
       else:
         # everything else (code, script, etc. that may contain < or >) should
         # be parsed until a closing tag
         # but: you may find <code> in <code>!?
         num = 1
         ie = i + 1
-        # print (html[ie:ie+20])
         while True:
           ie1 = html.find("</"+tb, ie)
           ie2 = html.find("<"+tb, ie)
           if ie1<0 and ie2<0:
-            # print ("Unclosed tag occured (%s, %s)" % (tb, ie))
+            if ie1<0: raise ValueError("Unclosed '<%s' element at position %s." % (tb, i))
             ie = len(html)
             break
-          # print ("c %s %s %s" % (ie, ie1, ie2))
           if ie1>=0 and (ie1<ie2 or ie2<0):
-            # print ("c1 %s %s" % (ie, ie1))
             num = num - 1
             ie = ie1 + len("</"+tb)
           if ie2>=0 and (ie2<ie1 or ie1<0):
-            # print ("c2 %s %s" % (ie, ie2))
             num = num + 1
             ie = ie2 + len("<"+tb)
           if num==0: break
@@ -284,7 +302,7 @@ class Degrotesque():
 
 
   # --- prettify
-  def prettify(self, html):
+  def prettify(self, html, useUnicode=False):
     """Prettifies (degrotesques) the given HTML snipplet using the given actions.
        It is assumed that the input is given in utf-8.
        The result is returned in utf-8 as well.
@@ -317,20 +335,21 @@ class Degrotesque():
         if a[0][1]!=None and closing==None:
           continue
         if closing!=None:
-          tmp = re.sub(a[0][1], a[1][1], html[ib:], 1)
+          if useUnicode: tmp = re.sub(a[0][1], a[2][1], html[ib:], 1)
+          else: tmp = re.sub(a[0][1], a[1][1], html[ib:], 1)
           l = closing.end() - closing.start() + len(tmp) - len(html[ib:]) 
           html = html[:ib] + tmp
-          marks = marks[:ib+closing.start()] + "0"*l + marks[ib+closing.end():]
+          marks = marks[:ib+closing.start()] + "1"*l + marks[ib+closing.end():]
           assert(len(html)==len(marks))
-        tmp = re.sub(a[0][0], a[1][0], html[i:], 1)
-        l = opening.end() - 0 + len(tmp) - len(html[i:])
+        if useUnicode: tmp = re.sub(a[0][0], a[2][0], html[i:], 1)
+        else: tmp = re.sub(a[0][0], a[1][0], html[i:], 1)
+        l = opening.end() + len(tmp) - len(html[i:])
         html = html[:i] + tmp
-        marks = marks[:i] + "0"*l + marks[i+opening.end():]
+        marks = marks[:i] + "1"*l + marks[i+opening.end():]
         assert(len(html)==len(marks))
-        i = i + opening.end() - 1
+        i = i + l - 1
         break
       i = i + 1
-    #print (html)
     return html
 
 
@@ -344,7 +363,8 @@ def getExtensions(extNames):
      extensions are used.
      Otherwise, the given string is split and returned as a list.
      :param extNames The names of extensions to process (or None if default 
-                     extensions shall be used)"""
+                     extensions shall be used)
+     :todo What about removing dots?"""
   if extNames==None or len(extNames)==0:
     return extensionsDB
   return [x.strip() for x in extNames.split(',')] 
@@ -402,20 +422,23 @@ def main(args):
   given, a backup of the original files is generated named as the original
   file with the appendix ".orig".
   """
+  sys.tracebacklimit = 0
   # parse options
-  optParser = OptionParser(usage="usage:\n  %prog [options]", version="%prog 0.6")
+  optParser = OptionParser(usage="usage:\n  degrotesque.py [options]", version="degrotesque.py 1.6")
   optParser.add_option("-i", "--input", dest="input", default=None, help="Defines files/folder to process")
-  optParser.add_option("-E", "--encoding", dest="encoding", default="utf-8", help="File encoding (default: 'utf-8'")
   optParser.add_option("-r", "--recursive", dest="recursive", action="store_true", default=False, help="Whether a given path shall be processed recursively")
   optParser.add_option("-B", "--no-backup", dest="no_backup", action="store_true", default=False, help="Whether no backup shall be generated")
+  optParser.add_option("-u", "--unicode", dest="unicode", action="store_true", default=False, help="Use unicode characters instead of HTML entities")
   optParser.add_option("-e", "--extensions", dest="extensions", default=None, help="Defines the extensions of files to process")
+  optParser.add_option("-E", "--encoding", dest="encoding", default="utf-8", help="File encoding (default: 'utf-8')")
   optParser.add_option("-s", "--skip", dest="skip", default=None, help="Defines the elements which contents shall not be changed")
   optParser.add_option("-a", "--actions", dest="actions", default=None, help="Defines the actions to perform")
   options, remaining_args = optParser.parse_args(args=args)
   # check options
   if options.input==None:
-    optParser.error("no input file(s) given...")
-    sys.exit()
+    print ("Error: no input file(s) given...", file=sys.stderr)
+    print ("Usage: degrotesque.py [options]+ -i <FILE>[,<FILE>]*", file=sys.stderr)
+    sys.exit(2)
   # setup degrotesque
   degrotesque = Degrotesque()
   degrotesque.setActions(options.actions)
@@ -431,9 +454,8 @@ def main(args):
       fd = io.open(f, mode="r", encoding=options.encoding)
       html = fd.read()
       fd.close()
-      #html = " <script> if(i<0) echo \"a\"</script> \"Hello World\" "
       # apply the beautifications
-      html = degrotesque.prettify(html)
+      html = degrotesque.prettify(html, options.unicode)
       # build a backup
       if not options.no_backup:
         shutil.copy(f, f+".orig")
@@ -442,7 +464,7 @@ def main(args):
       fd.write(html)
       fd.close()
     except ValueError as err:
-      print(str(err))
+      print (str(err))
       continue
 
 
